(*The length of the file to be compressed has to be less that 
[2^(size_max+1)-1].*)

let size_max = 62

let push_size_as_bits = Pack.push_int_as_bits size_max

let pop_bits_as_size = Pack.pop_bits_as_int size_max

(* Encoding. *)

let rec encode (dictionary : Huffman.encoding_dictionary) (text : Huffman.plain_text) : Huffman.cipher_text =
  lazy (
    match Lazy.force text with
    | MyStream.Nil ->
        MyStream.Nil
    | MyStream.Cons (symbol, text) ->
        Lazy.force (
          MyStream.concat
          (Huffman.encode_symbol dictionary symbol)
          (encode dictionary text)
        )
  )

(* This alternative definition of [encode] is equivalent to the above. *)
let encode dictionary text =
  MyStream.map_one_many
    (fun symbol accu -> 
      MyStream.concat (Huffman.encode_symbol dictionary symbol) accu)
    text

(* Decoding. *)

let rec decode (tree : Huffman.decoding_dictionary) (text : Huffman.cipher_text) : Huffman.plain_text =
  lazy (
    match Lazy.force text with
    | MyStream.Nil ->
        (* We have reached the end of the text. We are done. *)
        MyStream.Nil
    | MyStream.Cons (_, _) ->
        (* Decode one symbol and continue. Note that the recursive call to
           [decode] is in the scope of [lazy], so it is executed only when
           the symbol that follows [symbol] is demanded. *)
        let symbol, text = Huffman.decode_symbol tree text in
        MyStream.Cons (symbol, decode tree text)
  )

(* This alternative definition of [decode] is equivalent to the above. *)
let decode tree text =
  MyStream.map_many_one (Huffman.decode_symbol tree) text

let compress filename =
  (* Obtain the input -- a stream of 8-bit integers. *)
  let text = LazyIO.read_file filename in
  (* Build the dictionaries. This requires evaluating the input stream, hence
     reading the whole input file once. *)
  let size, encoding, decoding = Huffman.dictionaries_from_text text in
  (* Obtain the input again. If we re-used the old [text] here, the file would
     be read only once and the data would be kept in memory after we have
     built the dictionaries and while we encode the data. Instead, because we
     call [read_file] again, the file is read twice, and we work in constant
     space. Subtle! *)
  let text = LazyIO.read_file filename in
  (* Construct the output -- a stream of 8-bit integers. Note that not much
     work happens at this point: indeed, [Huffman.encode] runs in time O(1)
     and builds a suspension. *)
  let (encoded : bool MyStream.t) = encode encoding text in
  let (size_encoded : bool MyStream.t) =  
    push_size_as_bits size encoded in
  let (size_encoded : int MyStream.t) = 
    ByteStream.from_bits 
      size_encoded in
  let (decoding_size_encoded : int MyStream.t) = 
    Huffman.push_decoding_dictionary decoding (size_encoded) in
  (* Write the output file. *)
  LazyIO.write_file (filename ^ ".zz") decoding_size_encoded

let decompress filename =
  if not (Filename.check_suffix filename ".zz") then
    Printf.fprintf stderr "Error: the file name \"%s\" does not end in .zz. Ignored.\n%!" filename
  else begin
    (* On récupère le contenu d'un fichier compressé. Il contient dans l'ordre 
    les informations suivantes : le dictionnaire de 
    décodage, le nombre de symboles à décoder, et  
    le texte compressé par l'algorithme de Huffman.*)
    let (input : int MyStream.t) = LazyIO.read_file filename in
    (* On récupère le dictionnaire de décodage. *)
    let decoding, input = Huffman.pop_decoding_dictionary input in
    (* On convertit en flot binaire... *)
    let input = ByteStream.to_bits input in
    (* ...pour récupérer le nombre de symboles à décoder. *)
    let n, input = pop_bits_as_size input in
    let () = Printf.printf "packed integer n = %i\n" n in
    (* On décode le texte, donc [n] symboles (et pas plus, sans quoi nous 
    pourrions être en train de décoder le « padding », ce qui n'a aucun sens). 
    Une fois encore, on note qu'aucun calcul n'est effectué ici -- on construit 
    une suspension. *)
    let text = MyStream.take n (decode decoding input) in
    (* Le contenu du flot [text] est stocké dans le fichier de sortie.*)
    LazyIO.write_file (Filename.chop_suffix filename ".zz") text
  end
